import {
  DataRequestService
} from './../data-request/data-request.js';
import {
  _
} from '../../3rd.party.js';

let APP_CONFIGS = {};
/** 存储一个项目代码中存在的多个app的相关配置信息。
 * 结构是：
 * {
 *    'app1':{
 *      'base':{
 *          'appModulesConfig':{
 *              'module1':AppModuleConfig,
 *              'module2':AppModuleConfig,
 *          },
 *          'othersInfo':{...}
 *      },
 *      'module1':{...},
 *      'module2':{...}, 
 *    },
 * }
 * 其中base是app1自身的配置信息，而不是app1中某个功能模块的配置信息。
 * 暂时就app和module两级，如果要更深一级，结构也是类似，base也是自身配置信息，
 * 其余属性是相关功能的id值
 * 
 * @export
 */
const AppConfigService = (function () {

  function AppConfigService() {
    // 数据请求
    this.dataRequest = new DataRequestService();
  }

  /** 获取app自身的配置信息，不包含app中的功能模块配置信息
   * 
   * 
   * @params {string} appId 
   * @returnses {Object} 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.getAppBaseConfig = function (appId) {
    const appBaseConfig = this.getAppConfig(appId)['base'];
    return appBaseConfig ? appBaseConfig : {};
  };

  /** 获取app配置信息，包含base自身信息，以及相关的功能模块module信息
   * 
   * 
   * @params {string} appId 
   * @returnses {Object} 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.getAppConfig = function (appId) {
    return this.getAPPCONFIGS()[appId] ? this.getAPPCONFIGS()[appId] : {};
  };

  /** 从内存中获取appConfig信息
   * 
   * 
   * @returnss {Object} 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.getAPPCONFIGS = function () {
    return _.cloneDeep(APP_CONFIGS);
  };

  /** http请求app中的某个module信息
   * 
   * 
   * @params {string} appId 
   * @params {string} moduleId 
   * @returnses {Promise<any>} 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.getAppModuleConfig = function (appId, moduleId) {
    const moduleConfig = this.getAppConfig(appId)[moduleId];
    if (moduleConfig) {
      return new Promise((resolve, reject) => {
        resolve(moduleConfig);
      });
    } else {
      return this.getAppModuleConfigByHttp(appId, moduleId);
    }
  };

  /** http请求app中的某个module信息
   * 
   * 
   * @params {string} appId 
   * @params {string} moduleId 
   * @returnses {Promise<any>} 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.getAppModuleConfigByHttp = function (appId, moduleId) {
    const appModulesConfig = this.getAppBaseConfig(appId)['appModulesConfig'];
    if (appModulesConfig && appModulesConfig[moduleId]) {
      const uri = appModulesConfig[moduleId]['uri'];
      return this.dataRequest.getJsonDataMethodByPromise(uri).then(callback => {
        let {
          status,
          responseData
        } = callback;
        if (200 === status) {
          this.setAppModuleConfig(appId, moduleId, responseData);
          return responseData;
        } else {
          return responseData;
        }
      });
    } else {
      console.warn('当前应用没有配置功能模块的配置信息');
      return new Promise((resolve, reject) => {
        resolve(undefined);
      });
    }
  };

  /** 存储app自身的配置信息，不包含app中的功能模块配置信息
   * 
   * 
   * @params {string} appId 
   * @params {Object} baseConfig 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.setAppBaseConfig = function (appId, baseConfig) {
    const appConfig = this.getAppConfig(appId);
    appConfig['base'] = baseConfig;
    this.setAppConfig(appId, appConfig);
  };

  /** 存储app的信息，包含自身信息和相关的功能模块module信息
   * 
   * 
   * @params {string} appId 
   * @params {Object} appConfig 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.setAppConfig = function (appId, appConfig) {
    const appConfigs = this.getAPPCONFIGS();
    appConfigs[appId] = appConfig;
    this.setAPPCONFIGS(appConfigs);
  };

  /** 将appConfigs信息保存到内存对象中，并保持长久的生命周期
   * （直到网站被关闭才会释放相应的内容，清除相应的内容）
   * 
   * 
   * @params {Object} appConfigs 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.setAPPCONFIGS = function (appConfigs) {
    APP_CONFIGS = _.cloneDeep(appConfigs);
  };

  /** 将app中的某个module的配置信息存储起来
   * 
   * @params {string} appId 
   * @params {string} moduleId 
   * @params {Object} moduleConfig 
   * @memberof AppConfigService
   */
  AppConfigService.prototype.setAppModuleConfig = function (appId, moduleId, moduleConfig) {
    const appConfig = this.getAppConfig(appId);
    appConfig[moduleId] = moduleConfig;
    this.setAppConfig(appId, appConfig);
  };

  return AppConfigService;
}());

export {
  AppConfigService
};
