import {
  BaseToolService
} from './../base-tool/base-tool.js';
import {
  JsonProcessService
} from './../json-process/json-process.js';
import {
  axios
} from './../../3rd.party.js';
import {
  IpMappingService
} from '../ip-mapping/ip-mapping.js';

function DataRequestResponse(status, responseData) {
  // 结果状态码
  this.status = status;
  // 响应的数据
  this.responseData = responseData;
};

const DataRequestService = (function () {
  function DataRequestService() {
    this.baseTool = new BaseToolService();
    this.ipMapping = new IpMappingService();
    this.jsonProcess = new JsonProcessService();
    // 请求失败时重新发起请求的次数,
    this.retryCount = 5;
  }

  /** 检查option的数据响应类型responseType是否为指定的类型
   * 不是的话将其设置为指定的类型
   *
   * @params {*} option
   * @params {string} responseType
   * @returnses {*}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.checkOptionResponseType = function (option, responseType) {
    if (option) {
      if (!option.responseType || responseType !== option.responseType) {
        option.responseType = responseType;
      }
    } else {
      option = {
        responseType: responseType
      };
    }
    return option;
  };

  /** 以get的方式请求json格式的数据，结果以Promise的方式返回
   *
   *
   * @params {string} url 请求地址
   * @params {Object} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getJsonDataMethodByPromise = function (url, option, autoEncodeURI) {
    autoEncodeURI = this.baseTool.setPropDefaultValue(autoEncodeURI, true);
    option = this.checkOptionResponseType(option, 'json');
    return this.getMethodByPromise(url, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        if (data instanceof DataRequestResponse) {
          data.responseData = this.baseTool.convert2JsonObject(data.responseData);
        }
        return this.packageCallback(this.baseTool.convert2JsonObject(data), 200);
      }
    });
  };

  /**  以get的方式请求json格式的数据，
   * 并对结果进行处理，将可能存在的换行符、空格符替换掉，
   * 结果以Promise的方式返回
   *
   * @params {string} url 请求的url
   * @params {*} [option] 配置
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @params {boolean} [processNewLine] 是否处理换行符
   * @params {boolean} [processSpace] 是否处理空格符
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getProcessJsonDataMethodByPromise = function (url, option, autoEncodeURI, processNewLine, processSpace, ) {
    return this.getTextDataMethodByPromise(url, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        return this.packageCallback(this.jsonProcess.string2Json(data, processNewLine, processSpace), 200);
      }
    });
  };

  /** 以get的方式请求json数据。在发起请求时会对url地址进行ip地址映射，
   * 结果以Promise的方式返回
   *
   * @params {string} url 请求地址
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getJsonDataMethodByPromiseWithIpMapping = function (url, option, autoEncodeURI) {
    option = this.checkOptionResponseType(option, 'json');
    return this.getMethodByPromiseWithIpMapping(url, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        if (data instanceof DataRequestResponse) {
          data.responseData = this.baseTool.convert2JsonObject(data.responseData);
        }
        return this.packageCallback(data, 200);
      }
    });
  };

  /** 以get的方式请求json数据,并对请求得到的数据进行处理，
   * 将可能存在的换行符、空格符等处理掉。
   * 在发起请求时会对url地址进行ip地址映射，
   * 结果以Promise的方式返回
   *
   * @params {string} url 请求的url
   * @params {*} [option] 配置
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @params {boolean} [processNewLine] 是否处理换行符
   * @params {boolean} [processSpace] 是否处理空格符
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getProcessJsonDataMethodByPromiseWithIpMapping = function (url, option, autoEncodeURI, processNewLine, processSpace, ) {
    return this.getTextDataMethodByPromiseWithIpMapping(url, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        return this.packageCallback(this.jsonProcess.string2Json(data, processNewLine, processSpace), 200);
      }
    });
  };

  /** 以get的方式发起请求，结果以Promise的方式返回
   *
   *
   * @params {string} url 请求地址
   * @params {Object} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getMethodByPromise = function (url, option, autoEncodeURI) {
    autoEncodeURI = this.baseTool.setPropDefaultValue(autoEncodeURI, true);
    return axios.get(autoEncodeURI ? encodeURI(url) : url, option).then(data => {
      return this.packageCallback(data.request.response, 200);
    }).catch((e) => {
      if (this.retryCount > 0) {
        this.retryCount--;
        return this.getMethodByPromise(url, option, autoEncodeURI);
      } else {
        return this.handleError(e);
      }
    });
  };

  /**  以get的方式发起请求。在发起请求时会对url地址进行ip地址映射，
   * 结果以Promise的方式返回
   *
   * @params {string} url 请求地址
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getMethodByPromiseWithIpMapping = function (url, option, autoEncodeURI) {
    return this.ipMapping.mappingByPromise(url).then(ipMappingUrl => {
      return this.getMethodByPromise(ipMappingUrl, option, autoEncodeURI);
    }).catch((e) => {
      return this.packageCallback(e, 400);
    });
  };

  /** 以get的方式发起请求，返回text类型的结果，并以Promise的方式返回
   *
   *
   * @params {string} url
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getTextDataMethodByPromise = function (url, option, autoEncodeURI) {
    option = this.checkOptionResponseType(option, 'text');
    return this.getMethodByPromise(url, option, autoEncodeURI);
  };

  /** 以get的方式发起请求，同时对请求的url进行ip映射处理，
   * 返回text类型的结果，并以Promise的方式返回
   *
   * @params {string} url
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.getTextDataMethodByPromiseWithIpMapping = function (url, option, autoEncodeURI) {
    return this.ipMapping.mappingByPromise(url).then(ipMappingUrl => {
      return this.getTextDataMethodByPromise(ipMappingUrl, option, autoEncodeURI);
    }).catch((e) => {
      return this.packageCallback(e, 400);
    });
  };

  /** 错误信息处理
   *
   * @params {*} e
   * @returnses {*}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.handleError = function (e) {
    this.retryCount = 5;
    return this.packageCallback(e, 400);
  };

  /** 对结果进行打包
   *
   *
   * @params {*} callback 原先的返回值
   * @params {number} status 请求结果的状态码
   * @returnses {Object}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.packageCallback = function (callback, status) {
    if (callback instanceof DataRequestResponse) {
      return callback;
    }

    return new DataRequestResponse(status, callback);
  };

  /** 以post的方式请求json数据，结果以Promise的方式返回
   *
   * @params {string} url 请求地址
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postJsonDataMethodByPromise = function (url, body, option, autoEncodeURI) {
    option = this.checkOptionResponseType(option, 'json');
    return this.postMethodByPromise(url, body, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        if (data instanceof DataRequestResponse) {
          data.responseData = this.baseTool.convert2JsonObject(data.responseData);
        }
        return this.packageCallback(data, 200);
      }
    });
  };

  /** 以post的方式请求json数据,
   * 并对请求到的数据进行处理，将可能存在的换行符、空格符处理掉，
   * 结果以Promise的方式返回
   *
   * @params {string} url
   * @params {*} body
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @params {boolean} [processNewLine]
   * @params {boolean} [processSpace]
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postProcessJsonDataMethodByPromise = function (url, body, option, autoEncodeURI, processNewLine, processSpace) {
    return this.postTextDataMethodByPromise(url, body, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        return this.packageCallback(this.jsonProcess.string2Json(data, processNewLine, processSpace), 200);
      }
    });
  };

  /**  以post的方式请求json数据。在发起请求时会对url地址进行ip地址映射，
   * 结果以Promise的方式返回
   *
   * @params {string} url 请求地址
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postJsonDataMethodByPromiseWithIpMapping = function (url, body, option, autoEncodeURI) {
    option = this.checkOptionResponseType(option, 'json');
    return this.postMethodByPromiseWithIpMapping(url, body, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        if (data instanceof DataRequestResponse) {
          data.responseData = this.baseTool.convert2JsonObject(data.responseData);
        }
        return this.packageCallback(data, 200);
      }
    });
  };

  /** 以post的方式请求json数据，并对请求到的数据进行处理，将可能存在的换行符、空格符处理掉。
   * 在发起请求时会对url地址进行ip地址映射，
   * 结果以Promise的方式返回
   *
   * @params {string} url
   * @params {*} body
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @params {boolean} [processNewLine]
   * @params {boolean} [processSpace]
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postProcessJsonDataMethodByPromiseWithIpMapping = function (url, body, option, autoEncodeURI, processNewLine, processSpace, ) {
    return this.postTextDataMethodByPromiseWithIpMapping(url, body, option, autoEncodeURI).then(data => {
      if (data === undefined || data === null) {
        console.warn(`the ${url} request's callback is ${data}!`);
        return this.packageCallback(data, 400);
      } else {
        return this.packageCallback(this.jsonProcess.string2Json(data, processNewLine, processSpace), 200);
      }
    });
  };

  /** 以post的方式发起请求，结果以Promise的方式返回
   *
   *
   * @params {string} url 请求地址
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postMethodByPromise = function (url, body, option, autoEncodeURI) {
    autoEncodeURI = this.baseTool.setPropDefaultValue(autoEncodeURI, true);
    return axios.post(autoEncodeURI ? encodeURI(url) : url, body, option).then(data => {
      return this.packageCallback(data.request.response, 200);
    }).catch((e) => {
      if (this.retryCount > 0) {
        this.retryCount--;
        return this.postMethodByPromise(url, body, option, autoEncodeURI);
      } else {
        return this.handleError(e);
      }
    });
  };

  /** 以post的方式发起请求，请求text类型的数据，
   * 结果以Promise的方式返回
   *
   * @params {string} url
   * @params {*} body
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postTextDataMethodByPromise = function (url, body, option, autoEncodeURI) {
    option = this.checkOptionResponseType(option, 'text');
    return this.postMethodByPromise(url, body, option, autoEncodeURI);
  };

  /**  以post的方式发起请求。在发起请求时会对url地址进行ip地址映射，
   * 结果以Promise的方式返回
   *
   *
   * @params {string} url 请求地址
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postMethodByPromiseWithIpMapping = function (url, body, option, autoEncodeURI) {
    return this.ipMapping.mappingByPromise(url).then(ipMappingUrl => {
      return this.postMethodByPromise(ipMappingUrl, body, option, autoEncodeURI);
    }).catch((e) => {
      return this.handleError(e);
    });
  };

  /**  以post的方式发起请求,请求text类型的数据。在发起请求时会对url地址进行ip地址映射，
   * 结果以Promise的方式返回
   *
   * @params {string} url
   * @params {*} body
   * @params {*} [option]
   * @params {boolean} autoEncodeURI 是否对请求地址自动编码,默认是自动编码
   * @returnses {Promise<any>}
   * @memberof DataRequestService
   */
  DataRequestService.prototype.postTextDataMethodByPromiseWithIpMapping = function (url, body, option, autoEncodeURI) {
    option = this.checkOptionResponseType(option, 'text');
    return this.ipMapping.mappingByPromise(url).then(ipMappingUrl => {
      return this.postMethodByPromise(ipMappingUrl, body, option, autoEncodeURI);
    }).catch((e) => {
      this.handleError(e);
    });
  };

  /* 设置请求失败时，重新发起请求的次数
   */
  DataRequestService.prototype.setRetryCount = function (retryCount) {
    this.retryCount = retryCount;
  };

  return DataRequestService;
}());

export {
  DataRequestService,
  DataRequestResponse,
};
